/********************************************
mawk_execute.c

libmawk changes (C) 2009-2010, Tibor 'Igor2' Palinkas;
based on mawk code coming with the below copyright:

copyright 1991-1996, Michael D. Brennan

This is a source file for mawk, an implementation of
the AWK programming language.

Mawk is distributed without warranty under the terms of
the GNU General Public License, version 2, 1991.
********************************************/

#include "mawk.h"
#include "code.h"
#include "memory.h"
#include "symtype.h"
#include "field.h"
#include "bi_funct.h"
#include "bi_vars.h"
#include "regexp.h"
#include "repl.h"
#include "fin.h"
#include "debug.h"
#include <string.h>
#include "num.h"
#include "math_wrap.h"
#include "cell.h"

static int compare(mawk_state_t *, mawk_cell_t *);
static int d_to_index(mawk_state_t *, mawk_num_t);


/*  The stack machine that mawk_executes the code */

#ifdef	DEBUG
void DB_mawk_eval_overflow(mawk_state_t * MAWK)
{
	mawk_overflow(MAWK, "eval stack", EVAL_STACK_SIZE);
}
#endif

/* holds info for array loops (on a stack) */
typedef struct aloop_state {
	struct aloop_state *link;
	mawk_cell_t *var;										/* for(var in A) */
	mawk_string_t **base;
	mawk_string_t **ptr;
	mawk_string_t **limit;
} ALOOP_STATE;

/* clean up aloop stack on next, return, exit */
#define CLEAR_ALOOP_STACK() if(aloop_state){\
	    clear_aloop_stack(MAWK, aloop_state);\
	    aloop_state=(ALOOP_STATE*)0;}else

/* free the vector (aloop_state's ->base array and its members) */
static void aloop_free_vect(mawk_state_t *MAWK, ALOOP_STATE *top)
{
	while (top->ptr < top->limit) {
		free_STRING(*top->ptr);
		top->ptr++;
	}

	if (top->base != NULL)
		mawk_free(MAWK, top->base);
}


static void clear_aloop_stack(mawk_state_t *MAWK, ALOOP_STATE *top)
{
	ALOOP_STATE *q;
	do {
		aloop_free_vect(MAWK, top);
		q = top;
		top = q->link;
		MAWK_ZFREE(MAWK, q);
	} while (top);
}


/* Call awk function fbp with a_args already pushed on the stack. sp is
the stack pointer that is returned after the operation. */
mawk_cell_t *mawk_call(mawk_state_t * MAWK, mawk_cell_t * sp, FBLOCK * fbp, int a_args)
{
	mawk_cell_t *cp;
	int t;
	mawk_cell_t *nfp = sp - a_args + 1;	/* new fp for callee */
	mawk_cell_t *local_p = sp + 1;				/* first local argument on stack */
	char *type_p;									/* pts to type of an argument */
	SYMTAB *sym;
	void *old_userdata;

	if (fbp->code == NULL) {
		sym = mawk_find(MAWK, fbp->name, 1);
		if (sym->type == ST_C_FUNCTION) {
			sp[1].type = ST_NONE;
			old_userdata = MAWK->userdata;
			MAWK->userdata = sym->stval.c_function.userdata;
			sp = sym->stval.c_function.callback(MAWK, sp, a_args) + 1;
			MAWK->userdata = old_userdata;
		}
		else {
			mawk_rt_error(MAWK, "unresolved function call %s\n", fbp->name);
		}
		return sp;
	}
	else {
		if (fbp->nargs)
			type_p = fbp->typev + a_args - 1;

		/* create space for locals */
		t = fbp->nargs - a_args;		/* t is number of locals */
		while (t > 0) {
			t--;
			sp++;
			type_p++;
			sp->type = C_NOINIT;
			if (*type_p == ST_LOCAL_ARRAY)
				sp->ptr = (PTR) mawk_array_new(MAWK, NULL);
		}

		if (MAWK->debug_symbols)
			mawk_debug_callstack_push(MAWK, fbp);

		mawk_execute(MAWK, fbp->code, sp, nfp);

		if (MAWK->debug_symbols)
			mawk_debug_callstack_pop(MAWK);

	}

	/* cleanup the callee's arguments */
	/* putting return value at top of eval stack */
	if (sp >= nfp) {
		cp = sp + 1;								/* cp -> the function return */

		do {
			if (*type_p == ST_LOCAL_ARRAY) {
				if (sp >= local_p) {
					mawk_array_clear(MAWK, (mawk_array_t)(sp->ptr));
					MAWK_ZFREE(MAWK, (mawk_array_t) sp->ptr);
				}
			}
			else
				mawk_cell_destroy(MAWK, sp);
			type_p--;
			sp--;
		}
		while (sp >= nfp);

		mawk_cellcpy(MAWK, ++sp, cp);
		mawk_cell_destroy(MAWK, cp);
	}
	else
		sp++;												/* no arguments passed */
	return sp;
}

/******************* mawk_execute: the VM ********************/

/* cast cp to num and copy the value of the num to res_num (which is mawk_num_t) */
#define mawk_cast_get_num(res_num, cp) \
	do { \
		if (cp->type != C_NUM) \
			mawk_cast1_to_num(MAWK, cp); \
		res_num = cp->d.dval; \
	} while(0)


void mawk_execute(mawk_state_t *MAWK, register INST *cdp, register mawk_cell_t *sp, mawk_cell_t *fp)
{
	/* cdp -> code ptr, start execution here */
	/* sp -> eval_stack pointer */
	/* fp -> frame ptr into eval_stack for user defined functions */

	/* some useful temporaries */
	mawk_cell_t *cp;
	int t;
	mawk_num_t tmp_num;

	/* save state for array loops via a stack */
	ALOOP_STATE *aloop_state = (ALOOP_STATE *) 0;

	/* for moving the eval stack on deep recursion */
	mawk_cell_t *old_stack_base;
	mawk_cell_t *old_sp;

#ifdef	DEBUG
	mawk_cell_t *entry_sp = sp;
#endif


	if (fp) {
		/* we are a function call, check for deep recursion */
		if (sp > MAWK->stack_danger) {	/* change stacks */
			old_stack_base = MAWK->stack_base;
			old_sp = sp;
			MAWK->stack_base = (mawk_cell_t *) mawk_zmalloc(MAWK, sizeof(mawk_cell_t) * EVAL_STACK_SIZE);
			MAWK->stack_danger = MAWK->stack_base + DANGER;
			sp = MAWK->stack_base;
			/* waste 1 slot for ANSI, actually large model msdos breaks in
			   RET if we don't */
#ifdef	DEBUG
			entry_sp = sp;
#endif
		}
		else
			old_stack_base = (mawk_cell_t *) 0;
	}

	while (MAWK->rt_exit_code == 0)
		switch (cdp++->op) {

/* HALT only used by the disassemble now ; this remains
   so compilers don't offset the jump table */
		case _HALT:

		case _STOP:								/* only for range patterns */
#ifdef	DEBUG
			if (sp != entry_sp + 1)
				mawk_bozo(MAWK, "stop0");
#endif
			return;

		case _PUSHC:
			inc_sp();
			mawk_cellcpy(MAWK, sp, cdp++->ptr);
			break;

		case _PUSHD:
			inc_sp();
			sp->type = C_NUM;
			sp->d.dval = *(mawk_num_t *) cdp++->ptr;
			break;

		case _PUSHS:
			inc_sp();
			sp->type = C_STRING;
			sp->ptr = cdp++->ptr;
			string(sp)->ref_cnt++;
			break;

		case F_PUSHA:
			cp = (mawk_cell_t *) cdp->ptr;
			if (cp != MAWK->field) {
				if (MAWK->nf < 0)
					mawk_split_field0(MAWK);

				if (!(cp >= MAWK_NF && cp <= LAST_PFIELD)) {
					/* its a real field $1, $2 ...
					   If its greater than $NF, we have to
					   make sure its set to ""  so that
					   (++|--) and g?sub() work right
					 */
					t = mawk_field_addr_to_index(MAWK, cp);
					if (t > MAWK->nf) {
						mawk_cell_destroy(MAWK, cp);
						cp->type = C_STRING;
						cp->ptr = (PTR) & (MAWK->null_str);
						MAWK->null_str.ref_cnt++;
					}
				}
			}
			/* fall thru */

		case _PUSHA:
		case A_PUSHA:
			inc_sp();
			sp->type = C_NOINIT; /* normal varref, not C_ARR_REF */
			sp->ptr = cdp++->ptr;
			break;

		case _PUSHI:
			/* put contents of next address on stack */
			inc_sp();
			mawk_cellcpy(MAWK, sp, cdp++->ptr);
			break;

		case L_PUSHI:
			/* put the contents of a local var on stack,
			   cdp->op holds the offset from the frame pointer */
			inc_sp();
			mawk_cellcpy(MAWK, sp, fp + cdp++->op);
			break;

		case L_PUSHA:
			/* put a local address on eval stack */
			inc_sp();
			sp->type = C_NOINIT; /* normal varref, not C_ARR_REF */
			sp->ptr = (PTR) (fp + cdp++->op);
			break;


		case F_PUSHI:

			/* push contents of $i
			   cdp[0] holds & $i , cdp[1] holds i */

			inc_sp();
			if (MAWK->nf < 0)
				mawk_split_field0(MAWK);
			cp = (mawk_cell_t *) cdp->ptr;
			t = (cdp + 1)->op;
			cdp += 2;

			if (t <= MAWK->nf)
				mawk_cellcpy(MAWK, sp, cp);
			else {										/* an unset field */

				sp->type = C_STRING;
				sp->ptr = (PTR) & (MAWK->null_str);
				MAWK->null_str.ref_cnt++;
			}
			break;

		case NF_PUSHI:

			inc_sp();
			if (MAWK->nf < 0)
				mawk_split_field0(MAWK);
			mawk_cellcpy(MAWK, sp, MAWK_NF);
			break;

		case FE_PUSHA:

			if (sp->type != C_NUM)
				mawk_cast1_to_num(MAWK, sp);

			t = d_to_index(MAWK, sp->d.dval);
			if (t && MAWK->nf < 0)
				mawk_split_field0(MAWK);
			sp->ptr = (PTR) field_ptr(t);
			if (t > MAWK->nf) {
				/* make sure its set to "" */
				cp = (mawk_cell_t *) sp->ptr;
				mawk_cell_destroy(MAWK, cp);
				cp->type = C_STRING;
				cp->ptr = (PTR) & (MAWK->null_str);
				MAWK->null_str.ref_cnt++;
			}
			break;

		case FE_PUSHI:

			if (sp->type != C_NUM)
				mawk_cast1_to_num(MAWK, sp);

			t = d_to_index(MAWK, sp->d.dval);

			if (MAWK->nf < 0)
				mawk_split_field0(MAWK);
			if (t <= MAWK->nf)
				mawk_cellcpy(MAWK, sp, field_ptr(t));
			else {
				sp->type = C_STRING;
				sp->ptr = (PTR) & MAWK->null_str;
				MAWK->null_str.ref_cnt++;
			}
			break;


		case AE_PUSHA: /* global lookup and push array mawk_cell_t */
			/* top of stack has an expr, cdp->ptr points at an
			   array, replace the expr with the bifunct_target arr ref */
			cp = MAWK_ZMALLOC(MAWK, mawk_cell_t);
			mawk_cellcpy(MAWK, cp, sp);
			mawk_cell_destroy(MAWK, sp);
			sp->type = C_ARR_REF_BT;
			sp->ptr = (PTR) (mawk_array_t)(cdp->ptr);
			sp->d.idx_cell = cp;
			cdp++;
			break;

		case AE_PUSHA_WRARR: /* global lookup and push array mawk_cell_t _ref_ for a later write */
			{
				/* idx in sp, push array ptr to sp+1 */
				inc_sp();
				sp->ptr = (mawk_array_t)(cdp->ptr);
				cdp++;
				sp->type = C_ARR_REF;
			}
			break;


		case AE_PUSHI:
			/* top of stack has an expr, cdp->ptr points at an
			   array, replace the expr with the contents of the
			   cell inside the array */

			mawk_array_find(MAWK, (mawk_array_t)(cdp->ptr), sp, sp, MAWK_CREATE);
			cdp++;
			break;

		case LAE_PUSHI:
			/*  sp[0] is an expression
			   cdp->op is offset from frame pointer of a mawk_cell_t which
			   has an mawk_array_t in the ptr field, replace expr
			   with  array[expr]
			 */
			mawk_array_find(MAWK, (mawk_array_t)(fp[cdp->op].ptr), sp, sp, MAWK_CREATE);
			cdp++;
			break;

		case LAE_PUSHA: /* local array lookup and push array mawk_cell_t */
			/*  sp[0] is an expression
			   cdp->op is offset from frame pointer of a mawk_cell_t which
			   has an mawk_array_t in the ptr field, replace expr
			   with  bifunct_target arr ref
			 */
			cp = MAWK_ZMALLOC(MAWK, mawk_cell_t);
			mawk_cellcpy(MAWK, cp, sp);
			mawk_cell_destroy(MAWK, sp);
			sp->type = C_ARR_REF_BT;
			sp->ptr = (PTR) (mawk_array_t)(fp[cdp->op].ptr);
			sp->d.idx_cell = cp;
			cdp++;
			break;
		case LAE_PUSHA_WRARR: /* local array lookup and push array mawk_cell_t _ref_ for a later write */
			{
				/* idx in sp, push array ptr to sp+1 */
				inc_sp();
				sp->ptr = (mawk_array_t)(fp[cdp->op].ptr);
				cdp++;
				sp->type = C_ARR_REF;
			}
			break;


		case LA_PUSHA:
			/*  cdp->op is offset from frame pointer of a mawk_cell_t which
			   has an mawk_array_t in the ptr field. Push this ARRAY
			   on the eval stack
			 */
			inc_sp();
			sp->type = C_NOINIT; /* normal varref, not C_ARR_REF */
			sp->ptr = fp[cdp++->op].ptr;
			break;

		case SET_ALOOP:
			{
				ALOOP_STATE *ap = MAWK_ZMALLOC(MAWK, ALOOP_STATE);
				unsigned vector_size;

				ap->var = (mawk_cell_t *) sp[-1].ptr;
				ap->base = ap->ptr = mawk_array_loop_vector(MAWK, (mawk_array_t)(sp->ptr), &vector_size);
				ap->limit = ap->base + vector_size;
				sp -= 2;

				/* push onto aloop stack */
				ap->link = aloop_state;
				aloop_state = ap;
				cdp += cdp->op;
			}
			break;

		case ALOOP:
			{
				ALOOP_STATE *ap = aloop_state;
				if (ap->ptr < ap->limit) {
					mawk_cell_destroy(MAWK, ap->var);
					ap->var->type = C_STRING;
					ap->var->ptr = (PTR) * ap->ptr++;
					cdp += cdp->op;
				}
				else
					cdp++;
			}
			break;

		case POP_AL:
			{
				/* finish up an array loop */
				ALOOP_STATE *ap = aloop_state;
				aloop_state = ap->link;
				aloop_free_vect(MAWK, ap);
				MAWK_ZFREE(MAWK, ap);
			}
			break;

		case _POP:
			mawk_cell_destroy(MAWK, sp);
			sp--;
			break;

		case _ASSIGN:
			/* top of stack has an expr, next down is an
			   address, put the expression in *address and
			   replace the address with the expression */

			/* don't propagate type C_MBSTRN */
			if (sp->type == C_MBSTRN)
				mawk_check_strnum(MAWK, sp);
			sp--;
			mawk_cell_destroy(MAWK, ((mawk_cell_t *) sp->ptr));
			mawk_cellcpy(MAWK, sp->ptr, sp + 1);
			mawk_cellcpy(MAWK, sp, sp->ptr);
			mawk_cell_destroy(MAWK, sp + 1);
			break;

		case _ASSIGN_ARR:
			/* don't propagate type C_MBSTRN */
			if (sp->type == C_MBSTRN)
				mawk_check_strnum(MAWK, sp);
			sp-=2;
			/* sp is the index, sp+1 is the array ref and sp+2 is rvalue expr result */
			mawk_array_set_execute(MAWK, sp, sp+1, sp, sp+2);
			break;


		case F_ASSIGN:
			/* assign to a field  */
			if (sp->type == C_MBSTRN)
				mawk_check_strnum(MAWK, sp);
			sp--;
			mawk_field_assign(MAWK, (mawk_cell_t *) sp->ptr, sp + 1);
			mawk_cell_destroy(MAWK, sp + 1);
			mawk_cellcpy(MAWK, sp, (mawk_cell_t *) sp->ptr);
			break;

		case _ADD_ASG:
			if (sp->type != C_NUM)
				mawk_cast1_to_num(MAWK, sp);
			cp = (mawk_cell_t *) (sp - 1)->ptr;
			if (cp->type != C_NUM)
				mawk_cast1_to_num(MAWK, cp);
			P_nansafe1(cp->d.dval, (cp->d.dval + sp->d.dval), sp->d.dval);
			sp--;
			sp->type = C_NUM;
			sp->d.dval = cp->d.dval;
			break;

		case _ADD_ASG_ARR:
			/* convert the expression result */
			mawk_cast_get_num(tmp_num, sp);
			sp-=2;
			/* sp is the index, sp+1 is the array ref and sp+2 is rvalue expr result; after this sp+2 is the array val */
			if (!mawk_array_pure(MAWK, sp+1, 1)) {
				mawk_array_getnum_execute(MAWK, sp+2, sp+1, sp);
				P_nansafe1(sp[2].d.dval, (sp[2].d.dval + tmp_num), tmp_num);
				mawk_array_set_execute(MAWK, sp, sp+1, sp, sp+2);
			}
			else {
				mawk_array_getptr_execute(MAWK, cp, sp+1, sp);
				P_nansafe1(cp->d.dval, (cp->d.dval + tmp_num), tmp_num);
			}
			break;

		case _SUB_ASG:
			if (sp->type != C_NUM)
				mawk_cast1_to_num(MAWK, sp);
			cp = (mawk_cell_t *) (sp - 1)->ptr;
			if (cp->type != C_NUM)
				mawk_cast1_to_num(MAWK, cp);
			P_nansafe1(cp->d.dval, (cp->d.dval - sp->d.dval), sp->d.dval);
			sp--;
			sp->type = C_NUM;
			sp->d.dval = cp->d.dval;
			break;

		case _SUB_ASG_ARR:
			/* convert the expression result */
			mawk_cast_get_num(tmp_num, sp);
			sp-=2;
			/* sp is the index, sp+1 is the array ref and sp+2 is rvalue expr result; after this sp+2 is the array val */
			if (!mawk_array_pure(MAWK, sp+1, 1)) {
				mawk_array_getnum_execute(MAWK, sp+2, sp+1, sp);
				P_nansafe1(sp[2].d.dval, (sp[2].d.dval - tmp_num), tmp_num);
				mawk_array_set_execute(MAWK, sp, sp+1, sp, sp+2);
			}
			else {
				mawk_array_getptr_execute(MAWK, cp, sp+1, sp);
				P_nansafe1(cp->d.dval, (cp->d.dval - tmp_num), tmp_num);
			}
			break;

		case _MUL_ASG:
			if (sp->type != C_NUM)
				mawk_cast1_to_num(MAWK, sp);
			cp = (mawk_cell_t *) (sp - 1)->ptr;
			if (cp->type != C_NUM)
				mawk_cast1_to_num(MAWK, cp);
			P_nansafe1(cp->d.dval, (cp->d.dval * sp->d.dval), sp->d.dval);
			sp--;
			sp->type = C_NUM;
			sp->d.dval = cp->d.dval;
			break;

		case _MUL_ASG_ARR:
			/* convert the expression result */
			mawk_cast_get_num(tmp_num, sp);
			sp-=2;
			/* sp is the index, sp+1 is the array ref and sp+2 is rvalue expr result; after this sp+2 is the array val */
			if (!mawk_array_pure(MAWK, sp+1, 1)) {
				mawk_array_getnum_execute(MAWK, sp+2, sp+1, sp);
				P_nansafe1(sp[2].d.dval, (sp[2].d.dval * tmp_num), tmp_num);
				mawk_array_set_execute(MAWK, sp, sp+1, sp, sp+2);
			}
			else {
				mawk_array_getptr_execute(MAWK, cp, sp+1, sp);
				P_nansafe1(cp->d.dval, (cp->d.dval * tmp_num), tmp_num);
			}
			break;

		case _DIV_ASG:
			if (sp->type != C_NUM)
				mawk_cast1_to_num(MAWK, sp);
			cp = (mawk_cell_t *) (sp - 1)->ptr;
			if (cp->type != C_NUM)
				mawk_cast1_to_num(MAWK, cp);

#ifdef MAWK_NO_FLOAT
			{
				mawk_num_t d;
				d = sp--->d.dval;
				if (d != MAWK_NUM_ZERO)
					cp->d.dval /= d;
				else
					cp->d.dval = P_nan();
			}
#else
			{
				P_nansafe1(cp->d.dval, (cp->d.dval / sp->d.dval), sp->d.dval);
				sp--;
			}
#endif
			sp->type = C_NUM;
			sp->d.dval = cp->d.dval;
			break;


		case _DIV_ASG_ARR:
			/* convert the expression result */
			mawk_cast_get_num(tmp_num, sp);
			sp-=2;
			/* sp is the index, sp+1 is the array ref and sp+2 is rvalue expr result; after this sp+2 is the array val */
			if (!mawk_array_pure(MAWK, sp+1, 1)) {
				mawk_array_getnum_execute(MAWK, sp+2, sp+1, sp);

#ifdef MAWK_NO_FLOAT
				if (tmp_num != MAWK_NUM_ZERO)
					sp[2].d.dval /= tmp_num;
				else
					sp[2].d.dval = P_nan();
#else
				P_nansafe1(sp[2].d.dval, (sp[2].d.dval / tmp_num), tmp_num);
#endif
				mawk_array_set_execute(MAWK, sp, sp+1, sp, sp+2);
			}
			else {
				mawk_array_getptr_execute(MAWK, cp, sp+1, sp);
#ifdef MAWK_NO_FLOAT
				if (tmp_num != MAWK_NUM_ZERO)
					cp->d.dval /= tmp_num;
				else
					cp->d.dval = P_nan();
#else
				P_nansafe1(cp->d.dval, (cp->d.dval / tmp_num), tmp_num);
#endif
			}
			break;

		case _MOD_ASG:
			if (sp->type != C_NUM)
				mawk_cast1_to_num(MAWK, sp);
			cp = (mawk_cell_t *) (sp - 1)->ptr;
			if (cp->type != C_NUM)
				mawk_cast1_to_num(MAWK, cp);

			if (P_isnan_manual(cp->d.dval) || P_isnan_manual(sp->d.dval))
				cp->d.dval = P_nan();
			else {
#ifdef MAWK_NO_FLOAT
				cp->d.dval %= sp->d.dval;
#else
				cp->d.dval = P_fmod(cp->d.dval, sp->d.dval);
#endif
			}
			sp--;

			sp->type = C_NUM;
			sp->d.dval = cp->d.dval;
			break;

		case _MOD_ASG_ARR:
			/* convert the expression result */
			mawk_cast_get_num(tmp_num, sp);
			sp-=2;
			/* sp is the index, sp+1 is the array ref and sp+2 is rvalue expr result; after this sp+2 is the array val */
			if (!mawk_array_pure(MAWK, sp+1, 1)) {
				mawk_array_getnum_execute(MAWK, sp+2, sp+1, sp);

#ifdef MAWK_NO_FLOAT
				sp[2].d.dval %= tmp_num;
#else
				sp[2].d.dval = P_fmod(sp[2].d.dval, tmp_num);
#endif
				mawk_array_set_execute(MAWK, sp, sp+1, sp, sp+2);
			}
			else {
				mawk_array_getptr_execute(MAWK, cp, sp+1, sp);
#ifdef MAWK_NO_FLOAT
				cp->d.dval %= tmp_num;
#else
				cp->d.dval = P_fmod(cp->d.dval, tmp_num);
#endif
			}
			break;


		case _POW_ASG:
			if (sp->type != C_NUM)
				mawk_cast1_to_num(MAWK, sp);
			cp = (mawk_cell_t *) (sp - 1)->ptr;
			if (cp->type != C_NUM)
				mawk_cast1_to_num(MAWK, cp);
			cp->d.dval = mawk_num_pow(cp->d.dval, sp--->d.dval);
			sp->type = C_NUM;
			sp->d.dval = cp->d.dval;
			break;

		case _POW_ASG_ARR:
			/* convert the expression result */
			mawk_cast_get_num(tmp_num, sp);
			sp-=2;
			/* sp is the index, sp+1 is the array ref and sp+2 is rvalue expr result; after this sp+2 is the array val */
			if (!mawk_array_pure(MAWK, sp+1, 1)) {
				mawk_array_getnum_execute(MAWK, sp+2, sp+1, sp);
				sp[2].d.dval = mawk_num_pow(sp[2].d.dval, tmp_num);
				mawk_array_set_execute(MAWK, sp, sp+1, sp, sp+2);
			}
			else {
				mawk_array_getptr_execute(MAWK, cp, sp+1, sp);
				cp->d.dval = mawk_num_pow(cp->d.dval, tmp_num);
			}
			break;


		case F_ADD_ASG:
			if (sp->type != C_NUM)
				mawk_cast1_to_num(MAWK, sp);
			cp = (mawk_cell_t *) (sp - 1)->ptr;
			mawk_cellcpy(MAWK, &MAWK->tc, cp);
			mawk_cast1_to_num(MAWK, &MAWK->tc);
			P_nansafe1(MAWK->tc.d.dval, (MAWK->tc.d.dval + sp->d.dval), sp->d.dval);
			sp--;
			sp->type = C_NUM;
			sp->d.dval = MAWK->tc.d.dval;
			mawk_field_assign(MAWK, cp, &MAWK->tc);
			break;

		case F_SUB_ASG:
			if (sp->type != C_NUM)
				mawk_cast1_to_num(MAWK, sp);
			cp = (mawk_cell_t *) (sp - 1)->ptr;
			mawk_cellcpy(MAWK, &MAWK->tc, cp);
			mawk_cast1_to_num(MAWK, &MAWK->tc);
			P_nansafe1(MAWK->tc.d.dval, (MAWK->tc.d.dval - sp->d.dval), sp->d.dval);
			sp--;
			sp->type = C_NUM;
			sp->d.dval = MAWK->tc.d.dval;
			mawk_field_assign(MAWK, cp, &MAWK->tc);
			break;

		case F_MUL_ASG:
			if (sp->type != C_NUM)
				mawk_cast1_to_num(MAWK, sp);
			cp = (mawk_cell_t *) (sp - 1)->ptr;
			mawk_cellcpy(MAWK, &MAWK->tc, cp);
			mawk_cast1_to_num(MAWK, &MAWK->tc);

			P_nansafe1(MAWK->tc.d.dval, (MAWK->tc.d.dval * sp->d.dval), sp->d.dval);
			sp--;

			sp->type = C_NUM;
			sp->d.dval = MAWK->tc.d.dval;
			mawk_field_assign(MAWK, cp, &MAWK->tc);
			break;

		case F_DIV_ASG:
			if (sp->type != C_NUM)
				mawk_cast1_to_num(MAWK, sp);
			cp = (mawk_cell_t *) (sp - 1)->ptr;
			mawk_cellcpy(MAWK, &MAWK->tc, cp);
			mawk_cast1_to_num(MAWK, &MAWK->tc);

#ifdef MAWK_NO_FLOAT
		{
			mawk_num_t d;
			d = sp--->d.dval;
			if (d != MAWK_NUM_ZERO)
				MAWK->tc.d.dval /= d;
			else
				MAWK->tc.d.dval = P_nan();
		}
#else
			P_nansafe1(MAWK->tc.d.dval, (MAWK->tc.d.dval / sp->d.dval), sp->d.dval);
			sp--;
#endif
			sp->type = C_NUM;
			sp->d.dval = MAWK->tc.d.dval;
			mawk_field_assign(MAWK, cp, &MAWK->tc);
			break;

		case F_MOD_ASG:
			if (sp->type != C_NUM)
				mawk_cast1_to_num(MAWK, sp);
			cp = (mawk_cell_t *) (sp - 1)->ptr;
			mawk_cellcpy(MAWK, &MAWK->tc, cp);
			mawk_cast1_to_num(MAWK, &MAWK->tc);

#ifdef MAWK_NO_FLOAT
			{
				int d;
				d = sp--->d.dval;
				if (d != MAWK_NUM_ZERO)
					MAWK->tc.d.dval %= d;
				else
					MAWK->tc.d.dval = P_nan();
			}
#else
			MAWK->tc.d.dval = P_fmod(MAWK->tc.d.dval, sp--->d.dval);
#endif
			sp->type = C_NUM;
			sp->d.dval = MAWK->tc.d.dval;
			mawk_field_assign(MAWK, cp, &MAWK->tc);
			break;

		case F_POW_ASG:
			if (sp->type != C_NUM)
				mawk_cast1_to_num(MAWK, sp);
			cp = (mawk_cell_t *) (sp - 1)->ptr;
			mawk_cellcpy(MAWK, &MAWK->tc, cp);
			mawk_cast1_to_num(MAWK, &MAWK->tc);
			MAWK->tc.d.dval = mawk_num_pow(MAWK->tc.d.dval, sp--->d.dval);
			sp->type = C_NUM;
			sp->d.dval = MAWK->tc.d.dval;
			mawk_field_assign(MAWK, cp, &MAWK->tc);
			break;

		case _ADD:
			sp--;
			if (TEST2(sp) != TWO_NUMS)
				mawk_cast2_to_num(MAWK, sp);

			P_nansafe1(sp[0].d.dval, (sp[0].d.dval + sp[1].d.dval), sp[1].d.dval);
			break;

		case _SUB:
			sp--;
			if (TEST2(sp) != TWO_NUMS)
				mawk_cast2_to_num(MAWK, sp);

			P_nansafe1(sp[0].d.dval, (sp[0].d.dval - sp[1].d.dval), sp[1].d.dval);
			break;

		case _MUL:
			sp--;
			if (TEST2(sp) != TWO_NUMS)
				mawk_cast2_to_num(MAWK, sp);

			P_nansafe1(sp[0].d.dval, (sp[0].d.dval * sp[1].d.dval), sp[1].d.dval);
			break;

		case _DIV:
			sp--;
			if (TEST2(sp) != TWO_NUMS)
				mawk_cast2_to_num(MAWK, sp);

#ifdef MAWK_NO_FLOAT
			if (sp[1].d.dval != 0)
				sp[0].d.dval /= sp[1].d.dval;
			else
				sp[0].d.dval = P_nan();
#else
			P_nansafe1(sp[0].d.dval, (sp[0].d.dval / sp[1].d.dval), sp[1].d.dval);
#endif
			break;

		case _MOD:
			sp--;
			if (TEST2(sp) != TWO_NUMS)
				mawk_cast2_to_num(MAWK, sp);

#ifdef MAWK_NO_FLOAT
			{
				int d;
				d = sp[1].d.dval;
				if (d != MAWK_NUM_ZERO)
					sp[0].d.dval %= d;
				else
					sp[0].d.dval = P_nan();
			}
#else
			sp[0].d.dval = P_fmod(sp[0].d.dval, sp[1].d.dval);
#endif
			break;

		case _POW:
			sp--;
			if (TEST2(sp) != TWO_NUMS)
				mawk_cast2_to_num(MAWK, sp);
			sp[0].d.dval = mawk_num_pow(sp[0].d.dval, sp[1].d.dval);
			break;

		case _NOT:
			/* evaluates to 0.0 or 1.0 */
		reswitch_1:
			switch (sp->type) {
			case C_NOINIT:
				sp->d.dval = MAWK_NUM_ONE;
				break;
			case C_NUM:
				if (!P_isnan_manual(sp->d.dval))
					sp->d.dval = sp->d.dval != MAWK_NUM_ZERO ? MAWK_NUM_ZERO : MAWK_NUM_ONE;
				break;
			case C_STRING:
				sp->d.dval = string(sp)->len ? MAWK_NUM_ZERO : MAWK_NUM_ONE;
				free_STRING(string(sp));
				break;
			case C_STRNUM:						/* mawk_test as a number */
				sp->d.dval = sp->d.dval != MAWK_NUM_ZERO ? MAWK_NUM_ZERO : MAWK_NUM_ONE;
				free_STRING(string(sp));
				break;
			case C_MBSTRN:
				mawk_check_strnum(MAWK, sp);
				goto reswitch_1;
			default:
				mawk_bozo(MAWK, "bad type on eval stack");
			}
			sp->type = C_NUM;
			break;

		case _TEST:
			/* evaluates to 0.0 or 1.0 */
		reswitch_2:
			switch (sp->type) {
			case C_NOINIT:
				sp->d.dval = MAWK_NUM_ZERO;
				break;
			case C_NUM:
				if (!P_isnan_manual(sp->d.dval))
					sp->d.dval = sp->d.dval != MAWK_NUM_ZERO ? MAWK_NUM_ONE : MAWK_NUM_ZERO;
				break;
			case C_STRING:
				sp->d.dval = string(sp)->len ? MAWK_NUM_ONE : MAWK_NUM_ZERO;
				free_STRING(string(sp));
				break;
			case C_STRNUM:						/* mawk_test as a number */
				sp->d.dval = sp->d.dval != MAWK_NUM_ZERO ? MAWK_NUM_ONE : MAWK_NUM_ZERO;
				free_STRING(string(sp));
				break;
			case C_MBSTRN:
				mawk_check_strnum(MAWK, sp);
				goto reswitch_2;
			default:
				mawk_bozo(MAWK, "bad type on eval stack");
			}
			sp->type = C_NUM;
			break;

		case _UMINUS:
			if (sp->type != C_NUM)
				mawk_cast1_to_num(MAWK, sp);
			if (!P_isnan_manual(sp->d.dval))
				sp->d.dval = -sp->d.dval;
			break;

		case _UPLUS:
			if (sp->type != C_NUM)
				mawk_cast1_to_num(MAWK, sp);
			break;

		case _CAT:
			{
				unsigned len1, len2;
				char *str1, *str2;
				mawk_string_t *b;

				sp--;
				if (TEST2(sp) != TWO_STRINGS)
					mawk_cast2_to_str(MAWK, sp);
				str1 = string(sp)->str;
				len1 = string(sp)->len;
				str2 = string(sp + 1)->str;
				len2 = string(sp + 1)->len;

				b = mawk_new_STRING0(MAWK, len1 + len2);
				memcpy(b->str, str1, len1);
				memcpy(b->str + len1, str2, len2);
				free_STRING(string(sp));
				free_STRING(string(sp + 1));

				sp->ptr = (PTR) b;
				break;
			}

		case _PUSHINT:
			inc_sp();
			sp->type = cdp++->op;
			break;

		case _BUILTIN:
		case _PRINT:
			sp = (*(PF_CP) cdp++->ptr) (MAWK, sp);
			break;

		case _POST_INC:
			cp = (mawk_cell_t *) sp->ptr;
			if (cp->type != C_NUM)
				mawk_cast1_to_num(MAWK, cp);
			sp->type = C_NUM;
			sp->d.dval = cp->d.dval;
			if (!P_isnan_manual(cp->d.dval))
				cp->d.dval += MAWK_NUM_ONE;
			break;

		case _POST_INC_ARR:
			inc_sp();
			sp->type = C_NOINIT;
			sp-=2;
			/* sp is the index, sp+1 is the array ref and sp+2 is tmp; after this sp+2 is the array val */
			if (!mawk_array_pure(MAWK, sp+1, 1)) {
				mawk_array_getnum_execute(MAWK, sp+2, sp+1, sp);
				tmp_num = sp[2].d.dval;
				if (!P_isnan_manual(sp[2].d.dval))
					sp[2].d.dval += MAWK_NUM_ONE;
				mawk_array_set(MAWK, (mawk_array_t)sp[1].ptr, sp, sp+2);

			}
			else {
				mawk_array_getptr_execute(MAWK, cp, sp+1, sp);
				if (cp->type != C_NUM)
					mawk_cast1_to_num(MAWK, cp);
				tmp_num = cp->d.dval;
				if (!P_isnan_manual(cp))
					cp->d.dval += MAWK_NUM_ONE;
			}
			mawk_cell_destroy(MAWK, sp);
			sp->type = C_NUM;
			sp->d.dval = tmp_num;
			break;


		case _POST_DEC:
			cp = (mawk_cell_t *) sp->ptr;
			if (cp->type != C_NUM)
				mawk_cast1_to_num(MAWK, cp);
			sp->type = C_NUM;
			sp->d.dval = cp->d.dval;
			if (!P_isnan_manual(sp->d.dval))
				cp->d.dval -= MAWK_NUM_ONE;
			break;

		case _POST_DEC_ARR:
			inc_sp();
			sp->type = C_NOINIT;
			sp-=2;
			/* sp is the index, sp+1 is the array ref and sp+2 is tmp; after this sp+2 is the array val */
			if (!mawk_array_pure(MAWK, sp+1, 1)) {
				mawk_array_getnum_execute(MAWK, sp+2, sp+1, sp);
				tmp_num = sp[2].d.dval;
				if (!P_isnan_manual(sp[2].d.dval))
					sp[2].d.dval -= MAWK_NUM_ONE;
				mawk_array_set(MAWK, (mawk_array_t)sp[1].ptr, sp, sp+2);
			}
			else {
				mawk_array_getptr_execute(MAWK, cp, sp+1, sp);
				if (cp->type != C_NUM)
					mawk_cast1_to_num(MAWK, cp);
				tmp_num = cp->d.dval;
				if (!P_isnan_manual(cp))
					cp->d.dval -= MAWK_NUM_ONE;
			}
			mawk_cell_destroy(MAWK, sp);
			sp->type = C_NUM;
			sp->d.dval = tmp_num;
			break;

		case _PRE_INC:
			cp = (mawk_cell_t *) sp->ptr;
			if (cp->type != C_NUM)
				mawk_cast1_to_num(MAWK, cp);
			if (!P_isnan_manual(sp->d.dval))
				cp->d.dval += MAWK_NUM_ONE;
			sp->d.dval = cp->d.dval;
			sp->type = C_NUM;
			break;

		case _PRE_INC_ARR:
			inc_sp();
			sp->type = C_NOINIT;
			sp-=2;
			/* sp is the index, sp+1 is the array ref and sp+2 is tmp; after this sp+2 is the array val */
			if (!mawk_array_pure(MAWK, sp+1, 1)) {
				mawk_array_getnum_execute(MAWK, sp+2, sp+1, sp);
				if (!P_isnan_manual(sp[2].d.dval))
					sp[2].d.dval += MAWK_NUM_ONE;
				mawk_array_set(MAWK, (mawk_array_t)sp[1].ptr, sp, sp+2);
				mawk_cell_destroy(MAWK, sp);
				sp->type = C_NUM;
				sp->d.dval = sp[2].d.dval;
			}
			else {
				mawk_array_getptr_execute(MAWK, cp, sp+1, sp);
				if (cp->type != C_NUM)
					mawk_cast1_to_num(MAWK, cp);
				if (!P_isnan_manual(cp))
					cp->d.dval += MAWK_NUM_ONE;
				mawk_cell_destroy(MAWK, sp);
				sp->type = C_NUM;
				sp->d.dval = cp->d.dval;
			}
			break;


		case _PRE_DEC:
			cp = (mawk_cell_t *) sp->ptr;
			if (cp->type != C_NUM)
				mawk_cast1_to_num(MAWK, cp);
			if (!P_isnan_manual(sp->d.dval))
				cp->d.dval -= MAWK_NUM_ONE;
			sp->d.dval = cp->d.dval;
			sp->type = C_NUM;
			break;

		case _PRE_DEC_ARR:
			inc_sp();
			sp->type = C_NOINIT;
			sp-=2;
			/* sp is the index, sp+1 is the array ref and sp+2 is tmp; after this sp+2 is the array val */
			if (!mawk_array_pure(MAWK, sp+1, 1)) {
				mawk_array_getnum_execute(MAWK, sp+2, sp+1, sp);
				if (!P_isnan_manual(sp[2].d.dval))
					sp[2].d.dval -= MAWK_NUM_ONE;
				mawk_array_set(MAWK, (mawk_array_t)sp[1].ptr, sp, sp+2);
				mawk_cell_destroy(MAWK, sp);
				sp->type = C_NUM;
				sp->d.dval = sp[2].d.dval;
			}
			else {
				mawk_array_getptr_execute(MAWK, cp, sp+1, sp);
				if (cp->type != C_NUM)
					mawk_cast1_to_num(MAWK, cp);
				if (!P_isnan_manual(cp))
					cp->d.dval -= MAWK_NUM_ONE;
				mawk_cell_destroy(MAWK, sp);
				sp->type = C_NUM;
				sp->d.dval = cp->d.dval;
			}
			break;

		case F_POST_INC:
			cp = (mawk_cell_t *) sp->ptr;
			mawk_cellcpy(MAWK, &MAWK->tc, cp);
			mawk_cast1_to_num(MAWK, &MAWK->tc);
			sp->type = C_NUM;
			sp->d.dval = MAWK->tc.d.dval;
			if (!P_isnan_manual(MAWK->tc.d.dval))
				MAWK->tc.d.dval += MAWK_NUM_ONE;
			mawk_field_assign(MAWK, cp, &MAWK->tc);
			break;

		case F_POST_DEC:
			cp = (mawk_cell_t *) sp->ptr;
			mawk_cellcpy(MAWK, &MAWK->tc, cp);
			mawk_cast1_to_num(MAWK, &MAWK->tc);
			sp->type = C_NUM;
			sp->d.dval = MAWK->tc.d.dval;
			if (!P_isnan_manual(MAWK->tc.d.dval))
				MAWK->tc.d.dval -= MAWK_NUM_ONE;
			mawk_field_assign(MAWK, cp, &MAWK->tc);
			break;

		case F_PRE_INC:
			cp = (mawk_cell_t *) sp->ptr;
			mawk_cellcpy(MAWK, sp, cp);
			mawk_cast1_to_num(MAWK, sp);
			if (!P_isnan_manual(sp->d.dval))
				sp->d.dval += MAWK_NUM_ONE;
			mawk_field_assign(MAWK, cp, sp);
			break;

		case F_PRE_DEC:
			cp = (mawk_cell_t *) sp->ptr;
			mawk_cellcpy(MAWK, sp, cp);
			mawk_cast1_to_num(MAWK, sp);
			if (!P_isnan_manual(sp->d.dval))
				sp->d.dval -= MAWK_NUM_ONE;
			mawk_field_assign(MAWK, cp, sp);
			break;

		case _JMP:
			cdp += cdp->op;
			break;

		case _JNZ:
			/* jmp if top of stack is non-zero and pop stack */
			if (P_isnan(sp->d.dval))
				mawk_rt_error(MAWK, "NaN in conditional jump");
			if (mawk_test(MAWK, sp))
				cdp += cdp->op;
			else
				cdp++;
			mawk_cell_destroy(MAWK, sp);
			sp--;
			break;

		case _JZ:
			/* jmp if top of stack is zero and pop stack */
			if (P_isnan(sp->d.dval))
				mawk_rt_error(MAWK, "NaN in conditional jump");
			if (!mawk_test(MAWK, sp))
				cdp += cdp->op;
			else
				cdp++;
			mawk_cell_destroy(MAWK, sp);
			sp--;
			break;

		case _LJZ:
			/* special jump for logical and */
			/* this is always preceded by _TEST */
			if (P_isnan(sp->d.dval))
				mawk_rt_error(MAWK, "NaN in conditional jump");
			if (sp->d.dval == MAWK_NUM_ZERO) {
				/* take jump, but don't pop stack */
				cdp += cdp->op;
			}
			else {
				/* pop and don't jump */
				sp--;
				cdp++;
			}
			break;

		case _LJNZ:
			/* special jump for logical or */
			/* this is always preceded by _TEST */
			if (P_isnan(sp->d.dval))
				mawk_rt_error(MAWK, "NaN in conditional jump");
			if (sp->d.dval != MAWK_NUM_ZERO) {
				/* take jump, but don't pop stack */
				cdp += cdp->op;
			}
			else {
				/* pop and don't jump */
				sp--;
				cdp++;
			}
			break;

			/*  the relation operations */
			/*  compare() makes sure string ref counts are OK */
		case _EQ:
			t = compare(MAWK, --sp);
			sp->type = C_NUM;
				sp->d.dval = P_nansafe_exp1((t == 0 ? MAWK_NUM_ONE : MAWK_NUM_ZERO), t);
			break;

		case _NEQ:
			t = compare(MAWK, --sp);
			sp->type = C_NUM;
			sp->d.dval = P_nansafe_exp1((t ? MAWK_NUM_ONE : MAWK_NUM_ZERO), t);
			break;

		case _LT:
			t = compare(MAWK, --sp);
			sp->type = C_NUM;
			sp->d.dval = P_nansafe_exp1((t < 0 ? MAWK_NUM_ONE : MAWK_NUM_ZERO), t);
			break;

		case _LTE:
			t = compare(MAWK, --sp);
			sp->type = C_NUM;
			sp->d.dval = P_nansafe_exp1((t <= 0 ? MAWK_NUM_ONE : MAWK_NUM_ZERO), t);
			break;

		case _GT:
			t = compare(MAWK, --sp);
			sp->type = C_NUM;
			sp->d.dval = P_nansafe_exp1((t > 0 ? MAWK_NUM_ONE : MAWK_NUM_ZERO), t);
			break;

		case _GTE:
			t = compare(MAWK, --sp);
			sp->type = C_NUM;
			sp->d.dval = P_nansafe_exp1((t >= 0 ? MAWK_NUM_ONE : MAWK_NUM_ZERO), t);
			break;

		case _MATCH0:
			/* does $0 match, the RE at cdp? */

			inc_sp();
			if (MAWK->field->type >= C_STRING) {
				sp->type = C_NUM;
				sp->d.dval = mawk_REtest(MAWK, string(MAWK->field)->str, cdp++->ptr)
					? MAWK_NUM_ONE : MAWK_NUM_ZERO;

				break /* the case */ ;
			}
			else {
				mawk_cellcpy(MAWK, sp, MAWK->field);
				/* and FALL THRU */
			}

		case _MATCH1:
			/* does expr at sp[0] match RE at cdp */
			if (sp->type < C_STRING)
				mawk_cast1_to_str(MAWK, sp);
			t = mawk_REtest(MAWK, string(sp)->str, cdp++->ptr);
			free_STRING(string(sp));
			sp->type = C_NUM;
			sp->d.dval = t ? MAWK_NUM_ONE : MAWK_NUM_ZERO;
			break;


		case _MATCH2:
			/* does sp[-1] match sp[0] as re */
			mawk_cast_to_RE(MAWK, sp);

			if ((--sp)->type < C_STRING)
				mawk_cast1_to_str(MAWK, sp);
			t = mawk_REtest(MAWK, string(sp)->str, (sp + 1)->ptr);

			free_STRING(string(sp));
			sp->type = C_NUM;
			sp->d.dval = t ? MAWK_NUM_ONE : MAWK_NUM_ZERO;
			break;

		case A_TEST:
			/* entry :  sp[0].ptr-> an array
			   sp[-1]  is an expression

			   we compute (expression in array)  */
			sp--;
			t = mawk_array_find(MAWK, (mawk_array_t)((sp + 1)->ptr), sp, NULL, NO_MAWK_CREATE);
			mawk_cell_destroy(MAWK, sp);
			sp->type = C_NUM;
			sp->d.dval = t ? MAWK_NUM_ONE : MAWK_NUM_ZERO;
			break;

		case A_DEL:
			/* sp[0].ptr ->  array
			   sp[-1] is an expr
			   delete  array[expr]  */

			mawk_array_delete(MAWK, (mawk_array_t)(sp->ptr), sp - 1);
			mawk_cell_destroy(MAWK, sp - 1);
			sp -= 2;
			break;

		case DEL_A:
			/* free all the array at once */
			mawk_array_clear(MAWK, (mawk_array_t)(sp->ptr));
			sp--;
			break;

			/* form a multiple array index */
		case A_CAT:
			sp = mawk_array_cat(MAWK, sp, cdp->op);
			cdp++;
			break;

		case _EXIT:
			if (sp->type != C_NUM)
				mawk_cast1_to_num(MAWK, sp);
			MAWK->exit_code = d_to_i(sp->d.dval);
			sp--;
			/* fall thru */

		case _EXIT0:

			if (!MAWK->end_start)
				mawk_exit(MAWK, MAWK->exit_code);


			cdp = MAWK->end_start;
			MAWK->end_start = (INST *) 0;	/* makes sure next exit exits */

			if (MAWK->begin_start)
				mawk_zfree(MAWK, MAWK->begin_start, MAWK->begin_size);
			if (MAWK->main_start)
				mawk_zfree(MAWK, MAWK->main_start, MAWK->main_size);
			sp = MAWK->eval_stack - 1;	/* might be in user function */
			CLEAR_ALOOP_STACK();			/* ditto */
			break;

		case _JMAIN:								/* go from BEGIN code to MAIN code */
/*			mawk_zfree(MAWK, MAWK->begin_start, MAWK->begin_size); (causes double free!) */
			MAWK->begin_start = (INST *) 0;
			cdp = MAWK->main_start;
			if (MAWK->separate_begin)
				return;
			break;

		case _OMAIN:
			if (!MAWK->main_fin)
				mawk_FINopen_main(MAWK);
			if (!MAWK->main_fin)
				return;
			MAWK->restart_label = cdp;
			cdp = MAWK->next_label;
			break;

		case _NEXT:
			/* next might be inside an aloop -- clear stack */
			CLEAR_ALOOP_STACK();
			cdp = MAWK->next_label;
			break;

		case OL_GL:
			{
				char *p;
				unsigned len;

				p = mawk_FINgets(MAWK, MAWK->main_fin, &len);
				if (p == (void *) mawk_FIN_nomore)
					return;
				if (!p) {
					if (!MAWK->end_start)
						mawk_exit(MAWK, 0);

					cdp = MAWK->end_start;
/*					mawk_zfree(MAWK, MAWK->main_start, MAWK->main_size); (causes double free) */
					MAWK->main_start = MAWK->end_start = (INST *) 0;
				}
				else {
					mawk_set_field0(MAWK, p, len);
					cdp = MAWK->restart_label;
					MAWK->rt_nr++;
					MAWK->rt_fnr++;
				}
			}
			break;

			/* two kinds of OL_GL is a historical stupidity from working on
			   a machine with very slow floating point emulation */
		case OL_GL_NR:
			{
				char *p;
				unsigned len;
				p = mawk_FINgets(MAWK, MAWK->main_fin, &len);
				if (p == (void *) mawk_FIN_nomore)
					return;
				if (!p) {
					if (!MAWK->end_start)
						mawk_exit(MAWK, 0);

					cdp = MAWK->end_start;
					mawk_zfree(MAWK, MAWK->main_start, MAWK->main_size);
					MAWK->main_start = MAWK->end_start = (INST *) 0;
				}
				else {
					mawk_set_field0(MAWK, p, len);
					cdp = MAWK->restart_label;

					if (TEST2(NR) != TWO_NUMS) {
						mawk_cast2_to_num(MAWK, NR);
					}

					NR->d.dval += MAWK_NUM_ONE;
					MAWK->rt_nr++;
					FNR->d.dval += MAWK_NUM_ONE;
					MAWK->rt_fnr++;
				}
			}
			break;

		case _RANGE:
/* mawk_test a range pattern:  pat1, pat2 { action }
   entry :
       cdp[0].op -- a flag, mawk_test pat1 if on else pat2
       cdp[1].op -- offset of pat2 code from cdp
       cdp[2].op -- offset of action code from cdp
       cdp[3].op -- offset of code after the action from cdp
       cdp[4] -- start of pat1 code
*/

#define FLAG	cdp[0].op
#define PAT2	cdp[1].op
#define ACTION	  cdp[2].op
#define FOLLOW	  cdp[3].op
#define PAT1	  4

			if (FLAG) {								/* mawk_test again pat1 */
				mawk_execute(MAWK, cdp + PAT1, sp, fp);
				t = mawk_test(MAWK, sp + 1);
				mawk_cell_destroy(MAWK, sp + 1);
				if (t)
					FLAG = 0;
				else {
					cdp += FOLLOW;
					break;								/* break the switch */
				}
			}

			/* mawk_test against pat2 and then perform the action */
			mawk_execute(MAWK, cdp + PAT2, sp, fp);
			FLAG = mawk_test(MAWK, sp + 1);
			mawk_cell_destroy(MAWK, sp + 1);
			cdp += ACTION;
			break;

/* function calls  */

		case _RET0:
			inc_sp();
			sp->type = C_NOINIT;
			/* fall thru */

		case _RET:

#ifdef	DEBUG
			if (sp != entry_sp + 1)
				mawk_bozo(MAWK, "ret");
#endif
			if (old_stack_base) {			/* reset stack */
				/* move the return value */
				mawk_cellcpy(MAWK, old_sp + 1, sp);
				mawk_cell_destroy(MAWK, sp);
				mawk_zfree(MAWK, MAWK->stack_base, sizeof(mawk_cell_t) * EVAL_STACK_SIZE);
				MAWK->stack_base = old_stack_base;
				MAWK->stack_danger = old_stack_base + DANGER;
			}

			/* return might be inside an aloop -- clear stack */
			CLEAR_ALOOP_STACK();

			return;

		case _CALL:
			{
				/*  cdp[0] holds ptr to "function block"
				   cdp[1] holds number of input arguments
				 */
				FBLOCK *fbp = (FBLOCK *) cdp++->ptr;
				int a_args = cdp++->op;	/* actual number of args */

				sp = mawk_call(MAWK, sp, fbp, a_args);
			}
			break;
		case LOCATION:
			mawk_location_change(MAWK, cdp->op);
			cdp++;
			break;
		default:
			mawk_bozo(MAWK, "bad opcode");
		}
}


/*
  return 0 if a numeric is zero else return non-zero
  return 0 if a string is "" else return non-zero
*/
int mawk_test(mawk_state_t *MAWK, register mawk_cell_t *cp)
{
reswitch:

	switch (cp->type) {
	case C_NOINIT:
		return 0;
	case C_STRNUM:								/* mawk_test as a number */
	case C_NUM:
		return cp->d.dval != MAWK_NUM_ZERO;
	case C_STRING:
		return string(cp)->len;
	case C_MBSTRN:
		mawk_check_strnum(MAWK, cp);
		goto reswitch;
	default:
		mawk_bozo(MAWK, "bad cell type in call to mawk_test");
	}
	return 0;											/*can't get here: shutup */
}

/* compare cells at cp and cp+1 and
   frees STRINGs at those cells
*/
static int compare(mawk_state_t *MAWK, register mawk_cell_t *cp)
{
	int k;

reswitch:

	switch (TEST2(cp)) {
	case TWO_NOINITS:
		return 0;

	case TWO_NUMS:
	two_d:
		if (P_isnan_manual((cp + 1)->d.dval))
			return P_nan();
		if (P_isnan_manual(cp->d.dval))
			return P_nan();
		return cp->d.dval > (cp + 1)->d.dval ? 1 : cp->d.dval < (cp + 1)->d.dval ? -1 : 0;

	case TWO_STRINGS:
	case STRING_AND_STRNUM:
	two_s:
		k = strcmp(string(cp)->str, string(cp + 1)->str);
		free_STRING(string(cp));
		free_STRING(string(cp + 1));
		return k;

	case NOINIT_AND_NUM:
	case NOINIT_AND_STRNUM:
	case NUM_AND_STRNUM:
	case TWO_STRNUMS:
		mawk_cast2_to_num(MAWK, cp);
		goto two_d;
	case NOINIT_AND_STRING:
	case NUM_AND_STRING:
		mawk_cast2_to_str(MAWK, cp);
		goto two_s;
	case TWO_MBSTRNS:
		mawk_check_strnum(MAWK, cp);
		mawk_check_strnum(MAWK, cp + 1);
		goto reswitch;

	case NOINIT_AND_MBSTRN:
	case NUM_AND_MBSTRN:
	case STRING_AND_MBSTRN:
	case STRNUM_AND_MBSTRN:
		mawk_check_strnum(MAWK, cp->type == C_MBSTRN ? cp : cp + 1);
		goto reswitch;

	default:											/* there are no default cases */
		mawk_bozo(MAWK, "bad cell type passed to compare");
	}
	return 0;											/* shut up */
}

/* convert a number d to a field index	$d -> $i */
static int d_to_index(mawk_state_t *MAWK, mawk_num_t d)
{

	if (d > MAX_FIELD)
		mawk_rt_overflow(MAWK, "maximum number of fields", MAX_FIELD);

	if (d >= MAWK_NUM_ZERO)
		return (int) d;

	/* might include nan */
	mawk_rt_error(MAWK, "negative field index $%.6g", d);
	return 0;											/* shutup */
}
