/********************************************
files.c

libmawk changes (C) 2009-2013, Tibor 'Igor2' Palinkas;
based on mawk code coming with the below copyright:

copyright 1991-94.  Michael D. Brennan

This is a source file for mawk, an implementation of
the AWK programming language.

Mawk is distributed without warranty under the terms of
the GNU General Public License, version 2, 1991.
********************************************/
#ifndef _POSIX_SOURCE
/* needed on some systems to get strdup() */
#define _POSIX_SOURCE
#endif
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "mawk.h"
#include "files.h"
#include "memory.h"
#include "fin.h"
#include "vio.h"


/* mawk_find a file on file_list */
FILE_NODE *mawk_file_find(mawk_state_t *MAWK, mawk_string_t *sval, int type)
{
	register FILE_NODE *p = MAWK->file_list;
	FILE_NODE *q = (FILE_NODE *) 0;
	char *name = sval->str;
	vio_t *vf;

	while (1) {
		if (!p) {
			/* open a new one */
			p = MAWK_ZMALLOC(MAWK, FILE_NODE);

			p->vf = NULL;
			p->fin = NULL;

			switch (p->type = type) {
			case F_TRUNC:
				p->vf = mawk_vio_open(MAWK, name, MAWK_VIO_O_TRUNC);
				if (p->vf == NULL)
					goto out_failure;
				break;

			case F_APPEND:
				p->vf = mawk_vio_open(MAWK, name, MAWK_VIO_O_APPEND);
				if (p->vf == NULL)
					goto out_failure;
				break;

			case F_IN:
				if (!(p->fin = (PTR) mawk_FINopen(MAWK, name, 0))) {
					mawk_zfree(MAWK, p, sizeof(FILE_NODE));
					return (PTR) 0;
				}
				break;

			case PIPE_OUT:
			case PIPE_IN:
				vf = mawk_vio_open_pipe(MAWK, name, type);
				if (vf == NULL) {
					if (type == PIPE_OUT) {
						goto out_failure;
					}
					else {
						mawk_zfree(MAWK, p, sizeof(FILE_NODE));
						return (PTR) 0;
					}
				}
				if (type == PIPE_IN)
					p->fin = mawk_FIN_vio_open(MAWK, vf, 0);
				else
					p->vf = vf;
				break;

#ifdef	DEBUG
			default:
				mawk_bozo(MAWK, "bad file type");
#endif
			}
			/* successful open */
			p->name = sval;
			sval->ref_cnt++;
			break;										/* while loop */
		}

		/* search is by name and type */
		if ((strcmp(name, p->name->str) == 0 && ((p->type == type) ||
																						/* no distinction between F_APPEND and F_TRUNC here */
																						(p->type >= F_APPEND && type >= F_APPEND))))
		{
			/* found */
			if (!q)										/*at front of list */
				return p;
			/* delete from list for move to front */
			q->link = p->link;
			break;										/* while loop */
		}

		q = p;
		p = p->link;
	}															/* end while loop */

	/* put p at the front of the list */
	p->link = MAWK->file_list;
	return (MAWK->file_list = p);

out_failure:
	mawk_errmsg(MAWK, errno, "cannot open \"%s\" for output", name);
	mawk_exitval(MAWK, 2, NULL);
	return NULL;
}


/* Close a file and delete it's node from the file_list.
   Walk the whole list, in case a name has two nodes,
   e.g. < "/dev/tty" and > "/dev/tty"
*/
int mawk_file_close(mawk_state_t *MAWK, mawk_string_t *sval)
{
	FILE_NODE dummy;
	register FILE_NODE *p;
	FILE_NODE *q = &dummy;				/* trails p */
	FILE_NODE *hold;
	char *name = sval->str;
	int retval = -1;

	dummy.link = p = MAWK->file_list;
	while (p) {
		if (strcmp(name, p->name->str) == 0) {
			/* found */
			switch (p->type) {
				case F_TRUNC:
				case F_APPEND:
				case PIPE_OUT:
				case PIPE_IN:
					retval = mawk_vio_close(MAWK, p->vf);
					break;

				case F_IN:
#warning TODO: do this throug VIN?
					retval = mawk_vio_close(MAWK, p->vf);
					break;
			}
			free_STRING(p->name);
			hold = p;
			q->link = p = p->link;
			MAWK_ZFREE(MAWK, hold);
		}
		else {
			q = p;
			p = p->link;
		}
	}

	MAWK->file_list = dummy.link;
	return retval;
}

/*
mawk_find an output file with name == sval and fflush it
*/
int mawk_file_flush(mawk_state_t *MAWK, mawk_string_t *sval)
{
	int ret = -1;
	register FILE_NODE *p = MAWK->file_list;
	unsigned len = sval->len;
	char *str = sval->str;

	if (len == 0) {
		/* for consistency with gawk */
		mawk_flush_all_output(MAWK);
		return 0;
	}

	while (p) {
		if (IS_OUTPUT(p->type) && len == p->name->len && strcmp(str, p->name->str) == 0) {
			ret = 0;
			mawk_vio_flush(MAWK, p->vf);
			/* it's possible for a command and a file to have the same
			   name -- so keep looking */
		}
		p = p->link;
	}
	return ret;
}

void mawk_flush_all_output(mawk_state_t * MAWK)
{
	FILE_NODE *p;

	for (p = MAWK->file_list; p; p = p->link)
		if (IS_OUTPUT(p->type))
			mawk_vio_flush(MAWK, p->vf);
}


/* When we exit, we need to close and wait for all output pipes */

#if MAWK_NO_PIPE
void mawk_close_out_pipes(mawk_state_t * MAWK)
{
	if (!MAWK->do_exit)
		mawk_rt_error(MAWK, "pipes not supported");
}

#else

/* work around for bug in AIX 4.1 -- If there are exactly 16 or 
   32 or 48 ..., open files then the last one doesn't get flushed on
   exit.  So the following is now a misnomer as we'll really close
   all output.
*/

void mawk_close_out_pipes(mawk_state_t * MAWK)
{
	register FILE_NODE *p = MAWK->file_list;

	while (p) {
		if (IS_OUTPUT(p->type)) {
			if (mawk_vio_is_stdouts(MAWK, p->vf) || (!MAWK->dont_close_std))
				mawk_vio_close(MAWK, p->vf);
			if (p->type == PIPE_OUT)
				mawk_vio_close(MAWK, p->vf);
		}
		p = p->link;
	}
}

#endif /* MAWK_NO_PIPE */


