/********************************************
vio_orig.c - virtual IO: original mawk file/pipe IO implementation

libmawk changes (C) 2009-2013, Tibor 'Igor2' Palinkas;
based on mawk code coming with the below copyright:

copyright 1991-94.  Michael D. Brennan

This is a source file for mawk, an implementation of
the AWK programming language.

Mawk is distributed without warranty under the terms of
the GNU General Public License, version 2, 1991.
********************************************/
#ifndef _POSIX_SOURCE
/* needed on some systems to get strdup() */
#define _POSIX_SOURCE
#endif
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdarg.h>
#include "mawk.h"
#include "files.h"
#include "memory.h"
#include "fin.h"
#include "field.h"
#include "vio.h"

/* The original mawk file IO layer uses FILE * for output (because fprintf()
   and fflush() works on them) and int fd for input (probably for short reads
   and manual buffering). Pipes (background processes) also need a pid for
   wait().

   The vio reimplementation follows the same conventions, defining vio_t
   to contain a FILE *, an int fd and a pid. For output FILE * is required,
   input is done on int fd.
*/

extern int isatty(int);

#ifdef	V7
#include  <sgtty.h>							/* defines FIOCLEX */
#endif


#ifndef	 NO_FCNTL_H

#include <fcntl.h>
#define	 CLOSE_ON_EXEC(fd)    fcntl(fd, F_SETFD, 1)

#else
#define	 CLOSE_ON_EXEC(fd) ioctl(fd, FIOCLEX, (PTR) 0)
#endif


struct vio_s {
	pid_t pid; /* we need to wait() when we close a pipe */
	FILE *f;
	int fd;
};

static vio_t *vio_alloc_(void)
{
	vio_t *vf;
	vf = malloc(sizeof(vio_t));
	vf->pid = -1;
	vf->fd = -1;
	vf->f = NULL;
	return vf;
}

int mawk_vio_close(mawk_state_t *MAWK, vio_t *vf)
{
#warning TODO: temp hack until real vio instead of artificial buff
/* hack: if vf is NULL, we are doing the artificial buffering  */
	if (vf == NULL)
		return 0;

	if (vf->f != NULL) {
		fclose(vf->f);
		vf->f = NULL;
	}
	if (vf->fd >= 0) {
		close(vf->fd);
		vf->fd = -1;
	}
#ifndef MAWK_NO_FORK
	if (vf->pid > 0) {
		mawk_wait_for(MAWK, vf->pid);
		vf->pid = -1;
	}
#endif
	return 0;
}

int mawk_vio_flush(mawk_state_t *MAWK, vio_t *vf)
{
	if (vf->f != NULL) {
		if (fflush(vf->f) < 0) {
			mawk_errmsg(MAWK, errno, "unexpected write error");
			mawk_exitval(MAWK, 2, -1);
		}
	}
	return 0;
}

#warning TODO: this shall be more low level - move fnode allocation to somewhere else
void mawk_vio_setup_stdouts(mawk_state_t * MAWK)
{
	FILE_NODE *p, *q;

	if (MAWK->fnode_stdout == NULL) {
		p = MAWK_ZMALLOC(MAWK, FILE_NODE);
		p->link = (FILE_NODE *) 0;
		p->type = F_TRUNC;
		p->name = mawk_new_STRING(MAWK, "/dev/stdout");
		p->vf = vio_alloc_();
		p->vf->f = stdout;
		MAWK->fnode_stdout = p;
	}

	if (MAWK->fnode_stderr == NULL) {
		q = MAWK_ZMALLOC(MAWK, FILE_NODE);
		q->link = p;
		q->type = F_TRUNC;
		q->name = mawk_new_STRING(MAWK, "/dev/stderr");
		q->vf = vio_alloc_();
		q->vf->f = stderr;
		MAWK->file_list = q;
		MAWK->fnode_stderr = q;
	}
}

static void bufsetup(FILE *f)
{
	if (isatty(fileno(f)))
		setbuf(f, (char *) 0);
}

/* fopen() but no buffering to ttys */
static FILE *tfopen(const char *name, const char *mode)
{
	FILE *retval = fopen(name, mode);
	if (retval)
		bufsetup(retval);

	return retval;
}

#warning TODO print runtime warnings
vio_t *mawk_vio_open(mawk_state_t *MAWK, const char *name, mawk_vio_open_mode_t mode)
{
	vio_t *vf;
	FILE *f = NULL;
	int fd = -1;

	switch(mode) {
		case MAWK_VIO_O_TRUNC:
			f = tfopen(name, "w");
			if (f == NULL)
				return NULL;
			break;
		case MAWK_VIO_O_APPEND:
			f = tfopen(name, "a");
			if (f == NULL)
				return NULL;
			break;
		case MAWK_VIO_I:
			fd = open(name, O_RDONLY);
			if (fd < 0)
				return NULL;
	}

	vf = vio_alloc_();
	vf->f = f;
	vf->fd = fd;
	return vf;
}

vio_t *mawk_vio_fdopen(mawk_state_t *MAWK, int fd, mawk_vio_open_mode_t mode)
{
	vio_t *vf;
	FILE *f = NULL;

	switch(mode) {
		case MAWK_VIO_O_TRUNC:
			f = fdopen(fd, "w");
			if (f == NULL)
				return NULL;
			bufsetup(f);
			break;
		case MAWK_VIO_O_APPEND:
			f = fdopen(fd, "a");
			if (f == NULL)
				return NULL;
			bufsetup(f);
			break;
		case MAWK_VIO_I:
			break;
	}

	vf = vio_alloc_();
	vf->f = f;
	vf->fd = fd;
	return vf;
}

#ifdef MAWK_NO_FORK
vio_t *mawk_vio_open_pipe(mawk_state_t *MAWK, char *name, int type)
{
	if (!MAWK->do_exit)
		mawk_rt_error(MAWK, "process execution not supported");
	return NULL;
}

void mawk_vio_exec_shell(mawk_state_t *MAWK, char *cmd)
{
	if (!MAWK->do_exit)
		mawk_rt_error(MAWK, "process execution not supported");
}
#else

void mawk_vio_exec_shell(mawk_state_t *MAWK, char *cmd)
{
	char **envp = mawk_environ_extract(MAWK);
	char **s;
	if (envp == NULL) {
		mawk_errmsg(MAWK, errno, "failed to exec %s -c %s: can't set up environ[]", MAWK->shell, cmd);
		fflush(stderr);
		_exit(128);
	}
	execle(MAWK->shell, MAWK->shell, "-c", cmd, (char *) 0, envp);
	mawk_errmsg(MAWK, errno, "failed to exec %s -c %s", MAWK->shell, cmd);
	fflush(stderr);
	_exit(128);
}

#ifdef MAWK_NO_PIPE
vio_t *mawk_vio_open_pipe(mawk_state_t *MAWK, char *name, int type)
{
	if (!MAWK->do_exit)
		mawk_rt_error(MAWK, "pipes (and background processes) not supported");
	return NULL;
}
#else

vio_t *mawk_vio_open_pipe(mawk_state_t *MAWK, char *name, int type)
{
	int the_pipe[2], local_fd, remote_fd;
	vio_t *vf;
	pid_t pid;
	FILE *f;

	if (pipe(the_pipe) < 0)
		return NULL;

	local_fd = the_pipe[type == PIPE_OUT];
	remote_fd = the_pipe[type == PIPE_IN];

	if (type == PIPE_OUT) {
		f = fdopen(local_fd, "w");
		if (f == NULL) {
			close(local_fd);
			close(remote_fd);
			return NULL;
		}
	}

	/* to keep output ordered correctly */
	fflush(stdout);
	fflush(stderr);

	pid = fork();
	switch (pid) {
		case -1:
			close(local_fd);
			close(remote_fd);
			return NULL;
		case 0:
			close(local_fd);
			close(type == PIPE_IN);
			dup(remote_fd);
			close(remote_fd);
			mawk_vio_exec_shell(MAWK, name);
		default:
			close(remote_fd);
			/* we could deadlock if future child inherit the local fd ,
			   set close on exec flag */
#warning TODO: better do this by hand - close on exec is not really portable
			CLOSE_ON_EXEC(local_fd);
			break;
	}

	vf = vio_alloc_();
	vf->pid = pid;

	/* for an output pipe store FILE * for consistent writes */
	if (type == PIPE_OUT)
		vf->f = f;
	else
		vf->fd = local_fd;

	return vf;
}
#endif
#endif

int mawk_vio_putc(mawk_state_t *MAWK, vio_t *vf, char c)
{
	if (vf->f == NULL)
		return -1;
	if (fputc(c, vf->f) != c)
		return -1;
	return 0;
}

int mawk_vio_write(mawk_state_t *MAWK, vio_t *vf, void *data, int len)
{
	if (vf->f == NULL)
		return -1;
	if (fwrite(data, 1, len, vf->f) == 1)
		return 0;
	return  -1;
}

int mawk_vio_write_str(mawk_state_t *MAWK, vio_t *vf, const char *str)
{
	if (vf->f == NULL)
		return -1;
	return fprintf(vf->f, "%s", str);
}

int mawk_vio_printf(mawk_state_t *MAWK, vio_t *vf, const char *fmt, ...)
{
	int len;
	va_list ap;

	if (vf->f == NULL)
		return -1;

	va_start(ap, fmt);
	len = vfprintf(vf->f, fmt, ap);
	va_end(ap);
	return len;
}

int mawk_vio_error(mawk_state_t *MAWK, vio_t *vf)
{
	if (vf->f == NULL)
		return -1;
	return ferror(vf->f);
}

int mawk_vio_is_stdouts(mawk_state_t *MAWK, vio_t *vf)
{
	if (vf->f == NULL)
		return 0;
	return ((vf->f != stdout) && (vf->f != stderr));
}

int mawk_vio_setbuf(mawk_state_t *MAWK, vio_t *vf, int buf_enable)
{
	if (vf->f == NULL)
		return -1;

	if (buf_enable)
		setbuf(vf->f, "");
	else
		setbuf(vf->f, NULL);

	return 0;
}

int mawk_vio_read(mawk_state_t *MAWK, vio_t *vf, char *dst, long int size)
{
#warning TODO: temp hack until real vio instead of artificial buff
/* hack: if vf is NULL, we are doing the artificial buffering and there's no more to read */
	if (vf == NULL)
		return -2;

	if (vf->f != NULL)
		return -1;
	if (vf->fd < 0)
		return -1;
	return read(vf->fd, dst, size);
}


int mawk_vio_fgetsable(mawk_state_t *MAWK, vio_t *vf)
{
	if (vf == NULL)
		return 0;
	return vf->f != NULL;
}

char *mawk_vio_fgets(mawk_state_t *MAWK, vio_t *vf, char *s, int size)
{
	return fgets(s, size, vf->f);
}

void mawk_vio_fixstdin(mawk_state_t *MAWK, vio_t *vf)
{
	if ((isatty(vf->fd) && MAWK->rs_shadow.type == SEP_CHAR && MAWK->rs_shadow.c == '\n') || (MAWK->interactive_flag && vf->fd == 0)) {
		/* interactive, i.e., line buffer this file */
		if (vf->fd == 0)
			vf->f = stdin;
	}
}


/* original FIN code */
FIN *mawk_fin_fd0_open(mawk_state_t *MAWK, int main_flag)
{
	vio_t *vf = mawk_vio_fdopen(MAWK, 0, MAWK_VIO_I);
	if (vf == NULL)
		return NULL;

	return mawk_FIN_vio_open(MAWK, vf, main_flag);
}

vio_t *mawk_vio_finopen(mawk_state_t *MAWK, char *filename, int main_flag)
{
	if (filename[0] == '-' && filename[1] == 0)
		return mawk_vio_fdopen(MAWK, 0, main_flag);
	return mawk_vio_open(MAWK, filename, MAWK_VIO_I);
}
